module IO (
	IOMode(..),Handle,HandlePosn,SeekMode(..),
	stderr,stdin,stdout,
	openFile,hClose,
	hIsEOF,hGetChar,hGetLine,hGetContents,
	hPutChar,hPutStr,hPutStrLn,
	hGetPosn,hSetPosn,hSeek,
	isAlreadyExistsError,isDoesNotExistError,isPermissionError,
	isIllegalOperation,isFullError,isEOFError,isUserError,
	ioeGetErrorString,
	try,bracket,bracket_,
	module PreludeIO
  ) where

import PreludeIO
import PreludeInternalIOError
import PreludeInternal (string_to_char_list,char_list_to_string)
import _SystemArray
import StdFile,StdMisc

newtype Handle = Handle Int deriving Eq

data IOMode
  = ReadMode | WriteMode | AppendMode | ReadWriteMode
    deriving (Eq, Ord, Ix, Bounded, Enum, Read, Show)

:: HandlePosn = {hp_handle::!Handle,hp_posn::!Int}

data SeekMode
  = AbsoluteSeek | RelativeSeek | SeekFromEnd
    deriving (Eq, Ord, Ix, Bounded, Enum, Read, Show)

stderr :: Handle
stderr = file_to_handle stderr_file

stderr_file :: *File
stderr_file = `code inline {
	.d 0 0
	jsr	stderrF
	.o 0 2 f
 }

stdin :: Handle
stdin = file_to_handle stdio

stdout :: Handle
stdout = file_to_handle stdio

stdio :: *File
stdio = `code inline {
	.d 0 0
	jsr	stdioF
	.o 0 2 f
 }

file_to_handle :: !*File -> Handle 
file_to_handle file = `code inline {
	update_b 0 1
	pop_b 1
}

handle_to_file :: !Handle -> *File
handle_to_file handle = `code inline {
	pushI -1
	push_b 1
	update_b 1 2
	update_b 0 1
	pop_b 1
}

openFile :: FilePath -> IOMode -> IO Handle
openFile file_path io_mode
  = IO (openFileWorld file_path io_mode)

io_mode_to_file_mode ReadMode = FReadText
io_mode_to_file_mode WriteMode = FWriteText
io_mode_to_file_mode AppendMode = FAppendText
io_mode_to_file_mode ReadWriteMode = abort `"ReadWriteMode not yet implemented"

openFileWorld :: !FilePath -> !IOMode -> !*World -> (!Handle,!*World)
openFileWorld file_path io_mode world
  # file_path_cl = char_list_to_string file_path
    file_mode = io_mode_to_file_mode io_mode
    (ok,file,world) = fopen file_path_cl file_mode world
  | ok
  	= (file_to_handle file,world)
	# (IO io_error) = ioError PermissionError
	= io_error world

hClose :: Handle -> IO ()
hClose handle = IO (hCloseWorld handle)

hCloseWorld :: !Handle -> !*World -> (!(),!*World)
hCloseWorld handle world
  # (ok,world) = fclose (handle_to_file handle) world
  | ok
	= ((),world)
  	# (IO io_error) = ioError PermissionError
  	= io_error world

hIsEOF :: Handle -> IO Bool
hIsEOF h = IO (hIsEOFWorld h)

hIsEOFWorld :: !Handle -> !*World -> (!Bool,!*World)
hIsEOFWorld h world
  # file = handle_to_file h
    (eof,file) = fend file
  = (eof,world)

hGetChar :: Handle -> IO Char
hGetChar h = IO (hGetCharWorld h)

hGetCharWorld :: !Handle -> !*World -> (!Char,!*World)
hGetCharWorld h world
  # file = handle_to_file h
    (ok,c,file) = freadc file
  | ok
  	= (c,world)
  	# (error,file) = ferror file
	| not error
	  # (IO io_error) = ioError EOFError
	  = io_error world
	  # (IO io_error) = ioError PermissionError
	  = io_error world

hGetLine :: Handle -> IO [Char]
hGetLine h = IO (hGetLineWorld h)

hGetLineWorld :: !Handle -> !*World -> (![Char],!*World)
hGetLineWorld h world
  # file = handle_to_file h
    (line,file) = freadline file
    s = size line
  | s/=0
  	| line~[s-1]=='\n'
  	  = (string_except_last_char_to_char_list 1 line,world)
	  = (string_to_char_list line,world)
  	# (error,file) = ferror file
	| not error
	  # (IO io_error) = ioError EOFError
	  = io_error world
	  # (IO io_error) = ioError PermissionError
	  = io_error world
  where
	string_except_last_char_to_char_list :: Int -> {#Char} -> [Char]
	string_except_last_char_to_char_list i s
	  | i<size s
		#! c=s~[i-1]
		= c:string_except_last_char_to_char_list (i+1) s
		= []

hGetContents :: Handle -> IO [Char]
hGetContents h = IO (hGetContentsWorld h)

hGetContentsWorld :: Handle -> !*World -> (![Char],!*World)
hGetContentsWorld h world
  # file = handle_to_file h
    file = fshare file
  = (get_contents file,world)
  where
	get_contents file
	  # (ok,c,file) = sfreadc file
	  | ok
	  	= c:get_contents file
	  	= []

hPutChar :: Handle -> Char -> IO ()
hPutChar h c = IO (hPutCharWorld h c)

hPutCharWorld :: Handle -> Char -> !*World -> (!(),!*World)
hPutCharWorld h c world
  # file = handle_to_file h
    file = fwritec c file
    (error,file) = ferror file
  | not error
	= ((),world)
  	# (IO io_error) = ioError FullError
  	= io_error world

hPutStr :: Handle -> String -> IO ()
hPutStr h s = IO (hPutStrWorld h s)

hPutStrWorld :: Handle -> String -> !*World -> (!(),!*World)
hPutStrWorld h s world
  # file = handle_to_file h
    file = put_char_list s file
    (error,file) = ferror file
  | not error
	= ((),world)
  	# (IO io_error) = ioError FullError
  	= io_error world

hPutStrLn :: Handle -> String -> IO ()
hPutStrLn h s = IO (hPutStrLnWorld h s)

hPutStrLnWorld :: Handle -> String -> !*World -> (!(),!*World)
hPutStrLnWorld h s world
  # file = handle_to_file h
    file = put_char_list s file
    file = fwritec '\n' file
    (error,file) = ferror file
  | not error
	= ((),world)
  	# (IO io_error) = ioError FullError
  	= io_error world

put_char_list (c:cs) file
  = put_char_list cs (fwritec c file)
put_char_list [] file
  = file

hGetPosn :: Handle -> IO HandlePosn
hGetPosn h = IO (hGetPosnWorld h)

hGetPosnWorld :: !Handle -> !*World -> (!HandlePosn,!*World)
hGetPosnWorld h world
  # file = handle_to_file h
    (pos,file) = fposition file
  = ({hp_handle=file_to_handle file,hp_posn=pos},world)

hSetPosn :: HandlePosn -> IO ()
hSetPosn handle_posn = IO (hSetPosnWorld handle_posn)

hSetPosnWorld :: HandlePosn -> !*World -> (!(),!*World)
hSetPosnWorld {hp_handle,hp_posn} world
  # file = handle_to_file hp_handle
    (ok,file) = fseek file hp_posn FSeekSet
  | ok
	= ((),world)
  	# (IO io_error) = ioError PermissionError
  	= io_error world

hSeek :: Handle -> SeekMode -> Integer -> IO ()
hSeek h seek_mode pos = IO (hSeekWorld h seek_mode pos)

hSeekWorld :: !Handle -> !SeekMode -> !Integer -> !*World -> (!(),!*World)
hSeekWorld h seek_mode {integer_a,integer_s} world
  | size integer_a==0
	# file = handle_to_file h
 	  fseek_mode = seek_mode_to_fseek_mode seek_mode
	  (ok,file) = fseek file integer_s fseek_mode
 	| ok
	  = ((),world)
	  # (IO io_error) = ioError IllegalOperationError
	  = io_error world
	# (IO io_error) = ioError IllegalOperationError
	= io_error world

seek_mode_to_fseek_mode AbsoluteSeek = FSeekSet
seek_mode_to_fseek_mode RelativeSeek = FSeekCur
seek_mode_to_fseek_mode SeekFromEnd = FSeekEnd

--hPrint :: Show a => Handle -> a -> IO ()
--hPrint h v :== hPutStr h (show v)

isAlreadyExistsError :: !IOError -> Bool
isAlreadyExistsError AlreadyExistsError = True
isAlreadyExistsError _ = False

isDoesNotExistError :: !IOError -> Bool
isDoesNotExistError DoesNotExistError = True
isDoesNotExistError _ = False

isPermissionError :: !IOError -> Bool
isPermissionError PermissionError = True
isPermissionError _ = False

isEOFError :: !IOError -> Bool
isEOFError EOFError = True
isEOFError _ = False

isFullError :: !IOError -> Bool
isFullError FullError = True
isFullError _ = False

isIllegalOperation :: !IOError -> Bool
isIllegalOperation IllegalOperationError = True
isIllegalOperation _ = False

isUserError :: !IOError -> Bool
isUserError (UserError _) = True
isUserError _ = False

ioeGetErrorString :: !IOError -> String
ioeGetErrorString io_error
	= string_to_char_list (ioeGetError`String io_error)

try :: IO a -> IO (Either IOError a)
try f
  = catch
  	(do	r <- f
		return (Right r))
	(return . Left)

bracket :: IO a -> (a -> IO b) -> (a -> IO c) -> IO c
bracket before after m
  = do	x <- before
	rs <- try (m x)
	after x
	case rs of
	  Right r -> return r
	  Left e -> ioError e

bracket_ :: IO a -> (a -> IO b) -> IO c -> IO c
bracket_ before after m
  = do	x <- before
	rs <- try m
	after x
	case rs of
	  Right r -> return r
	  Left e -> ioError e

/*
instance Show Handle where ..
instance Eq HandlePosn where ..
instance Show HandlePosn where ..
data BufferMode = NoBuffering | LineBuffering | BlockBuffering (Maybe Int)
deriving (Eq, Ord, Read, Show)
hFileSize :: Handle -> IO Integer
isEOF :: IO Bool
isEOF = hIsEOF stdin
hSetBuffering :: Handle -> BufferMode -> IO ()
hGetBuffering :: Handle -> IO BufferMode
hFlush :: Handle -> IO ()
hWaitForInput :: Handle -> Int -> IO Bool
hReady :: Handle -> IO Bool
hReady h = hWaitForInput h 0
hLookAhead :: Handle -> IO Char
hPrint :: Show a => Handle -> a -> IO ()
hIsOpen :: Handle -> IO Bool
hIsClosed :: Handle -> IO Bool
hIsReadable :: Handle -> IO Bool
hIsWritable :: Handle -> IO Bool
hIsSeekable :: Handle -> IO Bool
isAlreadyInUseError :: IOError -> Bool
ioeGetHandle :: IOError -> Maybe Handle
ioeGetFileName :: IOError -> Maybe FilePath
*/
