implementation module Numeric where

from Ratio import numerator,denominator,%
from Array import ::Array,array,!
from Char import intToDigit,digitToInt,isDigit,isOctDigit,isHexDigit

from StdInt import class +,instance + Int,instance one Int
from StdMisc import abort

minExpt :== `0
maxExpt :== `1100

fromRat :: (RealFloat a) => !Rational -> a
fromRat x =
	if x == 0 then encodeFloat 0 0
	else if x < 0 then - fromRat' (-x)
	else fromRat' x
where
	fromRat' :: (RealFloat a) => Rational -> a
	fromRat' x = r where
		b = floatRadix r
		p = floatDigits r
		(minExp0, _) = floatRange r
		minExp = minExp0 - p
		xMin = toRational (expt b (p-1))
		xMax = toRational (expt b p)
		p0 = (integerLogBase b (numerator x) - integerLogBase b (denominator x) - p) `max` minExp
		f = if p0 < 0 then 1 % expt b (-p0) else expt b p0 % 1
		(x', p') = scaleRat (toRational b) minExp xMin xMax p0 (x / f)
		r = encodeFloat (round x') p'

	scaleRat :: Rational -> Int -> Rational -> Rational -> Int -> Rational -> (Rational, Int)
	scaleRat b minExp xMin xMax p x
		= if p <= minExp then
			(x, p)
		  else if x >= xMax then
			scaleRat b minExp xMin xMax (p+1) (x/b)
		  else if x < xMin then
			scaleRat b minExp xMin xMax (p-1) (x*b)
		  else
			(x, p)

	integerLogBase :: Integer -> Integer -> Int
	integerLogBase b i
		= if i < b then
			0
		  else
			let	l = 2 * integerLogBase (b*b) i
				doDiv :: Integer -> Int -> Int
				doDiv i l = if i < b then l else doDiv (i `div` b) (l+1)
			in doDiv (i `div` (b^l)) l

fromRat`Double :: !Rational -> Double
fromRat`Double r = fromRational r

fromRat`Float :: !Rational -> Float
fromRat`Float r = fromRational r

showSigned :: (Real a) => (a -> ShowS) -> Int -> a -> ShowS
showSigned showPos p x
	| x < 0
		= showParen (p > 6) (showChar '-' . showPos (-x))
		= showPos x

-- showInt, showOct, showHex are used for positive numbers only
showOct :: (Integral a) => a -> ShowS
showOct n = ShowIntAtBase 8 intToDigit n
showInt :: (Integral a) => a -> ShowS
showInt n = ShowIntAtBase 10 intToDigit n
showHex :: (Integral a) => a -> ShowS
showHex n = ShowIntAtBase 16 intToDigit n

showIntAtBase :: (Integral a) => a -> (Int -> Char) -> a -> [Char] -> [Char]
showIntAtBase base intToDig n rest
	= ShowIntAtBase base intToDig n rest

ShowIntAtBase base intToDig n rest
	:== showIntAtBase n rest
where
	showIntAtBase n rest
		| n < 0 = abort_showIntAtBase
		| n' == 0 = rest'
		| otherwise = showIntAtBase n' rest'
	where
		(n',d) = quotRem n base
		rest' = intToDig (fromIntegral d) : rest

abort_showIntAtBase 
	= abort `"Numeric.showIntAtBase: can't show negative numbers"

readSigned :: (Real a) => ReadS a -> ReadS a
readSigned readPos = readParen False read'
  where
  	read' r = read'' r ++
			[(-x,t) | ("-",s) <- lex r,
			 (x,t) <- read'' s]
	read'' r = [(n,s) | (str,s) <- lex r,
				(n,"") <- readPos str]

readInt :: (Integral a) => a -> (Char -> Bool) -> (Char -> Int) -> ![Char] -> [(a,[Char])]
readInt radix isDig digToInt s
	= ReadInt radix isDig digToInt s

ReadInt radix isDig digToInt s
	:== [(foldl1 (\n d -> n * radix + d) (map (fromIntegral . digToInt) ds), r)
		| (ds,r) <- nonnull isDig s ]

nonnull :: (Char -> Bool) -> ReadS String
nonnull p s = [(cs,t) | (cs@(_:_),t) <- [span p s]]

-- Unsigned readers for various bases
readDec :: (Integral a) => ReadS a
readDec = ReadInt 10 isDigit digitToInt
readOct :: (Integral a) => ReadS a
readOct = ReadInt 8 isOctDigit digitToInt
readHex :: (Integral a) => ReadS a
readHex = ReadInt 16 isHexDigit digitToInt

showEFloat :: (RealFloat a) => Maybe Int -> a -> ShowS
showEFloat d x = showString (formatRealFloat FFExponent d x)
showFFloat :: (RealFloat a) => Maybe Int -> a -> ShowS
showFFloat d x = showString (formatRealFloat FFFixed d x)
showGFloat :: (RealFloat a) => Maybe Int -> a -> ShowS
showGFloat d x = showString (formatRealFloat FFGeneric d x)
showFloat :: (RealFloat a) => a -> ShowS
showFloat x = showGFloat Nothing x

showEFloat`Double :: Maybe Int -> !Double -> ShowS
showEFloat`Double d x = showString (formatRealFloat`DoubleFloat FFExponent d x)
showFFloat`Double :: Maybe Int -> !Double -> ShowS
showFFloat`Double d x = showString (formatRealFloat`DoubleFloat FFFixed d x)
showGFloat`Double :: Maybe Int -> !Double -> ShowS
showGFloat`Double d x = showString (formatRealFloat`DoubleFloat FFGeneric d x)
showFloat`Double :: !Double -> ShowS
showFloat`Double x = showGFloat`Double Nothing x

showEFloat`Float :: Maybe Int -> !Float -> ShowS
showEFloat`Float d x = showString (formatRealFloat`DoubleFloat FFExponent d x)
showFFloat`Float :: Maybe Int -> !Float -> ShowS
showFFloat`Float d x = showString (formatRealFloat`DoubleFloat FFFixed d x)
showGFloat`Float :: Maybe Int -> !Float -> ShowS
showGFloat`Float d x = showString (formatRealFloat`DoubleFloat FFGeneric d x)
showFloat`Float :: !Float -> ShowS
showFloat`Float x = showGFloat`Float Nothing x

data FFFormat = FFExponent | FFFixed | FFGeneric

formatRealFloat :: (RealFloat a) => FFFormat -> Maybe Int -> a -> String
formatRealFloat fmt decs x
	= if isNaN x then
		"NaN"
	  else if isInfinite x then
		if x < 0 then "-Infinity" else "Infinity"
	  else if x < 0 || isNegativeZero x then
		'-' : doFmt fmt (floatToDigits 10 (-x))
	  else
		doFmt fmt (floatToDigits 10 x)
  where
	doFmt fmt (is, e)
		= let
			ds = map intToDigit is
		in
			case fmt of
				FFGeneric
					-> doFmt (if e < 0 || e > 7 then FFExponent else FFFixed) (is, e)
				FFExponent
					-> case decs of
							Nothing
								-> case ds of
									[] -> "0.0e0"
									[d] -> d : ".0e" ++ show (e-1)
									d:ds -> d : '.' : ds ++ 'e':show (e-1)
							Just dec
								-> let dec' = max dec 1 in
									case is of
										[] -> '0':'.':take dec' (repeat '0') ++ "e0"
										_
											-> let	(ei, is') = roundTo10 (dec'+1) is
													d:ds = map intToDigit
																(if ei > 0 then init is' else is')
												in d:'.':ds ++ "e" ++ show (e-1+ei)
				FFFixed
					-> case decs of
						Nothing
							| e > 0 -> take e (ds ++ repeat '0')
										++ '.' : mk0 (drop e ds)
							| otherwise -> "0." ++ mk0 (replicate (-e) '0' ++ ds)
						Just dec
							-> let dec' = max dec 0 in
								if e >= 0 then
									let	(ei, is') = roundTo10 (dec' + e) is
										(ls, rs) = splitAt (e+ei)
															(map intToDigit is')
									in mk0 ls ++ mkdot0 rs
								else
									let	(ei, is') = roundTo10 dec'
														(replicate (-e) 0 ++ is)
										d : ds = map intToDigit
													(if ei > 0 then is' else 0:is')
									in d : mkdot0 ds

formatRealFloat`DoubleFloat fmt decs x
	= if isNaN x then
		"NaN"
	  else if isInfinite x then
		if x < 0 then "-Infinity" else "Infinity"
	  else if x < 0 || isNegativeZero x then
		'-' : doFmt fmt (floatToDigits`DoubleFloat10 (-x))
	  else
		doFmt fmt (floatToDigits`DoubleFloat10 x)
  where
	doFmt fmt (is, e)
		= let
			ds = map intToDigit is
		in
			case fmt of
				FFGeneric
					-> doFmt (if e < 0 || e > 7 then FFExponent else FFFixed) (is, e)
				FFExponent
					-> case decs of
							Nothing
								-> case ds of
									[] -> "0.0e0"
									[d] -> d : ".0e" ++ show (e-1)
									d:ds -> d : '.' : ds ++ 'e':show (e-1)
							Just dec
								-> let dec' = max dec 1 in
									case is of
										[] -> '0':'.':take dec' (repeat '0') ++ "e0"
										_
											-> let	(ei, is') = roundTo10 (dec'+1) is
													d:ds = map intToDigit
																(if ei > 0 then init is' else is')
												in d:'.':ds ++ "e" ++ show (e-1+ei)
				FFFixed
					-> case decs of
						Nothing
							| e > 0 -> take e (ds ++ repeat '0')
										++ '.' : mk0 (drop e ds)
							| otherwise -> "0." ++ mk0 (replicate (-e) '0' ++ ds)
						Just dec
							-> let dec' = max dec 0 in
								if e >= 0 then
									let	(ei, is') = roundTo10 (dec' + e) is
										(ls, rs) = splitAt (e+ei)
															(map intToDigit is')
									in mk0 ls ++ mkdot0 rs
								else
									let	(ei, is') = roundTo10 dec'
														(replicate (-e) 0 ++ is)
										d : ds = map intToDigit
													(if ei > 0 then is' else 0:is')
									in d : mkdot0 ds

mk0 "" = "0"
mk0 s = s

mkdot0 "" = ""
mkdot0 s = '.' : s

roundTo10 :: Int -> [Int] -> (Int, [Int])
roundTo10 d is
	= case f d is of
		(0, is) -> (0, is)
		(1, is) -> (1, 1 : is)
  where
	b2 = 5

	f n [] = (0, replicate n 0)
	f 0 (i:_) = (if i >= b2 then 1 else 0, [])
	f d (i:is)
		= let	(c, ds) = f (d-1) is
				i' = c + i
		  in if i' == 10 then (1, 0:ds) else (0, i':ds)

floatToDigits :: (RealFloat a) => Integer -> a -> ([Int], Int)
floatToDigits _ 0
	= ([], 0)
floatToDigits base x
	= let
		(f0, e0) = decodeFloat x
		(minExp0, _) = floatRange x
		p = floatDigits x
		b = floatRadix x
		minExp = minExp0 - p
		(f, e)
			= let n = minExp - e0
			  in if n > 0 then
				  	((f0 `div` (b^n))::Integer, (e0+n)::Int)
				 else
				 	(f0::Integer, e0::Int)
		(r, s, mUp, mDn)
			= if e >= 0 then
				let be = b^e in
				if f == b^(p-1) then
					(f*be*b*2, 2*b, be*b, b)
				else
					(f*be*2, 2, be, be)
			  else
				if e > minExp && f == b^(p-1) then
					(f*b*2, b^(-e+1)*2, b, 1)
				else
					(f*2, b^(-e)*2, 1, 1)
		k = let	k0
					= if b==2 && base==10 then
						(p - 1 + e0) * 3 `div` 10
					  else
						ceiling ((log (fromInteger (f+1)) +
							fromIntegral e * log (fromInteger b)) /
							log (fromInteger base))
			
				fixup n
					= if n >= 0 then
						if r + mUp <= expt base n * s then n else fixup (n+1)
					  else
						if expt base (-n) * (r + mUp) <= s then n else fixup (n+1)
			in fixup k0

		gen :: [Int] -> Integer -> Integer -> Integer -> Integer -> [Int]
		gen ds rn sN mUpN mDnN
			= let
				(dnInteger, rn') = (rn * base) `divMod` sN
				dn = (fromIntegral dnInteger)::Int
				mUpN' = mUpN * base
				mDnN' = mDnN * base
			  in case (rn' < mDnN', rn' + mUpN' > sN) of
				(True, False) -> dn : ds
				(False, True) -> dn+1 : ds
				(True, True) -> if rn' * 2 < sN then dn : ds else dn+1 : ds
				(False, False) -> gen (dn:ds) rn' sN mUpN' mDnN'
		rds
			= if k >= 0 then
				gen [] r (s * expt base k) mUp mDn
			  else
				let bk = expt base (-k)
				in gen [] (r * bk) s (mUp * bk) (mDn * bk)
	  in (reverse rds, k)

floatToDigits`Double :: Integer -> !Double -> ([Int], Int)
floatToDigits`Double base x
	= floatToDigits`DoubleFloat base x

floatToDigits`Float :: Integer -> !Float -> ([Int], Int)
floatToDigits`Float base x
	= floatToDigits`DoubleFloat base x

floatToDigits`DoubleFloat _ 0
	= ([], 0)
floatToDigits`DoubleFloat base x
	= let
		(f0, e0) = decodeFloat x
		minExp = -1074
		(f, e)
			= let n = minExp - e0
			  in if n > 0 then
				  	((f0 `div` (2^n))::Integer, (e0+n)::Int)
				 else
				 	(f0::Integer, e0::Int)
		(r, s, mUp, mDn)
			= if e >= 0 then
				let be = 2^e in
				if f == 2^52 then
					(f*be*4, 4, be*2, 2)
				else
					(f*be*2, 2, be, be)
			  else
				if e > minExp && f == 2^52 then
					(f*4, 2^(-e+1)*2, 2, 1)
				else
					(f*2, 2^(-e)*2, 1, 1)
		k = let	k0
					= if base==10 then
						(52 + e0) * 3 `div` 10
					  else
						ceiling ((log (fromInteger (f+1)) +
							fromIntegral e * log 2) /
							log (fromInteger base))
			
				fixup n
					= if n >= 0 then
						if r + mUp <= expt base n * s then n else fixup (n+1)
					  else
						if expt base (-n) * (r + mUp) <= s then n else fixup (n+1)
			in fixup k0

		gen :: [Int] -> Integer -> Integer -> Integer -> Integer -> [Int]
		gen ds rn sN mUpN mDnN
			= let
				(dnInteger, rn') = (rn * base) `divMod` sN
				dn = (fromIntegral dnInteger)::Int
				mUpN' = mUpN * base
				mDnN' = mDnN * base
			  in case (rn' < mDnN', rn' + mUpN' > sN) of
				(True, False) -> dn : ds
				(False, True) -> dn+1 : ds
				(True, True) -> if rn' * 2 < sN then dn : ds else dn+1 : ds
				(False, False) -> gen (dn:ds) rn' sN mUpN' mDnN'
		rds
			= if k >= 0 then
				gen [] r (s * expt base k) mUp mDn
			  else
				let bk = expt base (-k)
				in gen [] (r * bk) s (mUp * bk) (mDn * bk)
	  in (reverse rds, k)

floatToDigits`DoubleFloat10 0
	= ([], 0)
floatToDigits`DoubleFloat10 x
	= let
		(f0, e0) = decodeFloat x
		minExp = -1074
		(f, e)
			= let n = minExp - e0
			  in if n > 0 then
				  	((f0 `div` (2^n))::Integer, (e0+n)::Int)
				 else
				 	(f0::Integer, e0::Int)
		(r, s, mUp, mDn)
			= if e >= 0 then
				let be = 2^e in
				if f == 2^52 then
					(f*be*4, 4, be*2, 2)
				else
					(f*be*2, 2, be, be)
			  else
				if e > minExp && f == 2^52 then
					(f*4, 2^(-e+1)*2, 2, 1)
				else
					(f*2, 2^(-e)*2, 1, 1)
		k = let	k0 = (52 + e0) * 3 `div` 10
				fixup n
					= if n >= 0 then
						if r + mUp <= (10 ^ n) * s then n else fixup (n+1)
					  else
						if (10 ^ (-n)) * (r + mUp) <= s then n else fixup (n+1)
			in fixup k0

		gen10 :: [Int] -> Integer -> Integer -> Integer -> Integer -> [Int]
		gen10 ds rn sN mUpN mDnN
			= let
				(dnInteger, rn') = (rn * 10) `divMod` sN
				dn = (fromIntegral dnInteger)::Int
				mUpN' = mUpN * 10
				mDnN' = mDnN * 10
			  in case (rn' < mDnN', rn' + mUpN' > sN) of
				(True, False) -> dn : ds
				(False, True) -> dn+1 : ds
				(True, True) -> if rn' * 2 < sN then dn : ds else dn+1 : ds
				(False, False) -> gen10 (dn:ds) rn' sN mUpN' mDnN'
		rds
			= if k >= 0 then
				gen10 [] r (s * (10 ^ k)) mUp mDn
			  else
				let bk = 10 ^ (-k)
				in gen10 [] (r * bk) s (mUp * bk) (mDn * bk)
	  in (reverse rds, k)

/*
from StdDebug import trace_tn
from StdString import
	class +++(..),instance +++ {#Char},
	class toString(..),instance toString {#Char},instance toString Int
from PreludeIntegerToString import integer_to_string
*/

expt :: Integer -> Int -> Integer
expt base n
	= if base == 2 && n >= minExpt && n <= maxExpt then
		expts!n
	  else
		base^n

expts :: Array Int Integer
expts =: array (minExpt,maxExpt) [(n,2^n) | n <- [minExpt .. maxExpt]]

readFloat :: ![Char] -> [(a,[Char])] | RealFrac a
readFloat r
	= [	(fromRational ((n%1)*10^^(k-d)),t)
	  | (n,d,s) <- readFix r,
		(k,t) <- readExp s] ++
		[ (0/0, t) | ("NaN",t) <- lex r] ++
		[ (1/0, t) | ("Infinity",t) <- lex r]
  where
	readFix r = [(read (ds++ds'), length ds', t)
				| (ds,d) <- lexDigits r,
				  (ds',t) <- lexFrac d ]

	lexFrac ('.':ds) = lexDigits ds
	lexFrac s = [("",s)]

	readExp (e:s) | e `elem` "eE" = readExp' s
	readExp s = [(0,s)]
	readExp' ('-':s) = [(-k,t) | (k,t) <- readDec s]
	readExp' ('+':s) = readDec s
	readExp' s = readDec s

lexDigits :: ![Char] -> [([Char],[Char])]
lexDigits s = [(cs,t) | (cs@(_:_),t) <- [span isDigit s]]
